#include <Wire.h>
#include <Adafruit_Sensor.h>
#include <Adafruit_BNO055.h>
#include <utility/imumaths.h>
#include <Arduino.h>
#include <SPI.h>
#include <math.h>

#include <Nextion.h>
#include <stdio.h>
#include <stdlib.h>
#include <driver/i2s.h>
#include <SD.h>
#include <esp_task_wdt.h>

//  Global Variables/objects    
static const i2s_port_t i2s_num = I2S_NUM_0; // i2s port number
File WavFile; // File pointer to current music filebeing played
float speedFactorcurrent = 1; // indicates the current ratio between song's original bpm and gait bpm
float speedFactorprev = 0; // used to determine whether there has been a change in speedFactor
uint32_t BytesReadSoFarMusicTime = 0; // keeps track of musical progress for switching todifferent speed versions of song
String gnrstr; // rck, pop, rap, or lat
String targetSPMstr; // 060, 075, 090, or 105
String currentSpeedstr; // 080, 085, 090, 095, 100, 110, or 120 
String WavFileName; // Object for root of SD card directory, concatenation of targetSPMStr, gnrstr, and currentSpeedstr
int volume = 50; // volume on a scale of 0 to 100

boolean start = false;
float current_SPM;
float target_SPM;
int genre;
//------------------------------------------------------------------------------------------------------------------------
 
//    SD Card
          #define SD_CS         10         // SD Card chip select
   
//    I2S
          #define I2S_DOUT      4          // i2S Data out oin
          #define I2S_BCLK      41         // Bit clock
          #define I2S_LRC       40         // Left/Right clock, also known as Frame clock or word select
          #define I2S_NUM       0          // i2s port number

// Wav File reading
          #define NUM_BYTES_TO_READ_FROM_FILE 1024    // How many bytes to read from wav file at a time

//------------------------------------------------------------------------------------------------------------------------

//------------------------------------------------------------------------------------------------------------------------
// I2S configuration structures
 
static const i2s_config_t i2s_config = {
    .mode = (i2s_mode_t)(I2S_MODE_MASTER | I2S_MODE_TX),
    .sample_rate = 44100,
    .bits_per_sample = I2S_BITS_PER_SAMPLE_16BIT,
    .channel_format = I2S_CHANNEL_FMT_RIGHT_LEFT, // use stereo wav files
    .communication_format = (i2s_comm_format_t)(I2S_COMM_FORMAT_I2S | I2S_COMM_FORMAT_I2S_MSB),
    .intr_alloc_flags = ESP_INTR_FLAG_LEVEL1,       // high interrupt priority
    .dma_buf_count = 8,                             // 8 buffers
    .dma_buf_len = 1024,                            // 1K per buffer, so 8K of buffer space
    .use_apll=0,
    .tx_desc_auto_clear= true, 
    .fixed_mclk=-1    
};
 
// These are the physical wiring connections to our I2S decoder board/chip from the esp32, there are other connections
// required for the chips mentioned at the top (but not to the ESP32), please visit the page mentioned at the top for
// further information regarding these other connections.
 
static const i2s_pin_config_t pin_config = 
      {
          .bck_io_num = I2S_BCLK,                           // The bit clock connectiom, goes to pin 27 of ESP32
          .ws_io_num = I2S_LRC,                             // Word select, also known as word select or left right clock
          .data_out_num = I2S_DOUT,                         // Data out from the ESP32, connect to DIN on 38357A
          .data_in_num = I2S_PIN_NO_CHANGE                  // we are not interested in I2S data into the ESP32
      };

struct WavHeader_Struct
      {
          //   RIFF Section    
          char RIFFSectionID[4];      // Letters "RIFF"
          uint32_t Size;              // Size of entire file less 8
          char RiffFormat[4];         // Letters "WAVE"
          
          //   Format Section    
          char FormatSectionID[4];    // letters "fmt"
          uint32_t FormatSize;        // Size of format section less 8
          uint16_t FormatID;          // 1=uncompressed PCM
          uint16_t NumChannels;       // 1=mono,2=stereo
          uint32_t SampleRate;        // 44100, 16000, 8000 etc.
          uint32_t ByteRate;          // =SampleRate * Channels * (BitsPerSample/8)
          uint16_t BlockAlign;        // =Channels * (BitsPerSample/8)
          uint16_t BitsPerSample;     // 8,16,24 or 32
        
          // Data Section
          char DataSectionID[4];      // The letters "data"
          uint32_t DataSize;          // Size of the data that follows
      }WavHeader;



//------------------------------------------------------------------------------------------------------------------------

// A boolean test to validate the recieved wavfile data
bool ValidWavData(WavHeader_Struct* Wav)
{
  
  if(memcmp(Wav->RIFFSectionID,"RIFF",4)!=0) 
  {    
    // Serial.print("Invalid data - Not RIFF format");
    return false;        
  }
  if(memcmp(Wav->RiffFormat,"WAVE",4)!=0)
  {
    // Serial.print("Invalid data - Not Wave file");
    return false;           
  }
  if(memcmp(Wav->FormatSectionID,"fmt",3)!=0) 
  {
    // Serial.print("Invalid data - No format section found");
    return false;       
  }
  if(memcmp(Wav->DataSectionID,"data",4)!=0) 
  {
    // Serial.print("Invalid data - data section not found");
    return false;      
  }
  if(Wav->FormatID!=1) 
  {
    // Serial.print("Invalid data - format Id must be 1");
    return false;                          
  }
  if(Wav->FormatSize!=16) 
  {
    // Serial.print("Invalid data - format section size must be 16.");
    return false;                          
  }
  if((Wav->NumChannels!=1)&(Wav->NumChannels!=2))
  {
    // Serial.print("Invalid data - only mono or stereo permitted.");
    return false;   
  }
  if(Wav->SampleRate>48000) 
  {
    // Serial.print("Invalid data - Sample rate cannot be greater than 48000");
    return false;                       
  }
  if((Wav->BitsPerSample!=8)& (Wav->BitsPerSample!=16)) 
  {
    // Serial.print("Invalid data - Only 8 or 16 bits per sample permitted.");
    return false;                        
  }
  return true;
}

// Function that clears the memory arrays of the SD Card.
void clearSD()
{
  byte sd = 0;
  digitalWrite(SS, LOW);
  while (sd != 255)
  {
    sd = SPI.transfer(255);
    // Serial.print("sd=");
    // Serial.println(sd);
  }
  digitalWrite(SS, HIGH);
}

void SDCardInit()
{        
  SPI.begin();
  delay(10);
  boolean b;
  // *** SD Card **
  b = SD.begin(SS);
  if (!b)
  {
    delay(100);
    clearSD();
    delay(100);
    b = SD.begin(SS);
  }
  if (b)
    ;
    // Serial.println("SD Card started.");
  else
    // Serial.println("SD Card failed to start!");
    // Serial.println();  

    pinMode(SD_CS, OUTPUT); 
    digitalWrite(SD_CS, LOW);
    delay(100);
    digitalWrite(SD_CS, HIGH); // SD card chips select
    if(!SD.begin(SD_CS))
    {
        // Serial.println("Error talking to SD card!");
        while(true);                  // end program
    }
    
}


bool FillI2SBuffer(byte* Samples,uint16_t BytesInBuffer)
{
    // Writes bytes to buffer, returns true if all bytes sent else false, keeps track itself of how many left
    // to write, so just keep calling this routine until returns true to know they've all been written, then
    // you can re-fill the buffer
    
    size_t BytesWritten;                        // Returned by the I2S write routine, 
    static uint16_t BufferIdx=0;                // Current pos of buffer to output next
    uint8_t* DataPtr;                           // Point to next data to send to I2S
    uint16_t BytesToSend;                       // Number of bytes to send to I2S
    
    // To make the code eaier to understand I'm using to variables to some calculations, normally I'd write this calcs
    // directly into the line of code where they belong, but this make it easier to understand what's happening
    
    DataPtr=Samples+BufferIdx;                               // Set address to next byte in buffer to send out
    BytesToSend=BytesInBuffer-BufferIdx;                     // This is amount to send (total less what we've already sent)
    i2s_write(i2s_num,DataPtr,BytesToSend,&BytesWritten,1);  // Send the bytes, wait 1 RTOS tick to complete
    BufferIdx+=BytesWritten;                                 // increasue by number of bytes actually written
    
    if(BufferIdx>=BytesInBuffer)                 
    {
      // sent out all bytes in buffer, reset and return true to indicate this
      BufferIdx=0; 
      return true;                             
    }
    else
      return false;       // Still more data to send to I2S so return false to indicate this
}

uint16_t ReadFile(byte* Samples)
{
    static uint32_t BytesReadSoFar=0;                   // Number of bytes read from file so far
    uint16_t BytesToRead;                               // Number of bytes to read from the file
    uint16_t i;                                         // loop counter
    int16_t SignedSample;                               // Single Signed Sample        
    
    if(BytesReadSoFar+NUM_BYTES_TO_READ_FROM_FILE>WavHeader.DataSize)   // If next read will go past the end then adjust the 
      BytesToRead=WavHeader.DataSize-BytesReadSoFar;                    // amount to read to whatever is remaining to read
    else
      BytesToRead=NUM_BYTES_TO_READ_FROM_FILE;                          // Default to max to read
      
    WavFile.read(Samples,BytesToRead);                  // Read in the bytes from the file
    BytesReadSoFar+=BytesToRead;                        
    // Update the total bytes read in so far
    // NEW
    BytesReadSoFarMusicTime+=(uint32_t)( (float)BytesToRead / speedFactorcurrent);
    
    if(BytesReadSoFar>=WavHeader.DataSize)              // Have we read in all the data?
    {
      WavFile.seek(44);                                 // Reset to start of wav data  
      BytesReadSoFar=0;                                 // Clear to no bytes read in so far  
      BytesReadSoFarMusicTime = 0;                     
    }
    // Change Volume
    for(i=0;i<BytesToRead;i+=2){
        *((int16_t *)(Samples+i))=(*((int16_t *)(Samples+i)))*volume/100; 
    }
    //}

    return BytesToRead;                                 // return the number of bytes read into buffer
}

// switch WavFileName str based on the current speed factor
void switchWavFileName(){ 
  if(speedFactorcurrent > 1.15){
    currentSpeedstr = String("120");
  } 
  else if(speedFactorcurrent > 1.05){
    currentSpeedstr = String("110");
  }
  else if(speedFactorcurrent > 0.95){
    currentSpeedstr = String("100");
  }
  else if(speedFactorcurrent > 0.85){
    currentSpeedstr = String("090");
  }
  else{
    currentSpeedstr = String("080");
  }

  WavFileName = "/" + targetSPMstr + gnrstr + currentSpeedstr + ".wav";
}

void PlayWav()
{
  // Serial.println("Supposed to be playing");
  static bool ReadingFile=true;                       // True if reading file from SD. false if filling I2S buffer
  static byte Samples[NUM_BYTES_TO_READ_FROM_FILE];   // Memory allocated to store the data read in from the wav file
  static uint16_t BytesRead;                          // Num bytes actually read from the wav file which will either be
                                                      // NUM_BYTES_TO_READ_FROM_FILE or less than this if we are very
                                                      // near the end of the file. i.e. we can't read beyond the file.
  // New functions
  if (speedFactorcurrent == speedFactorprev){ //if there is no change in speedFactor, no change needed
  }else{ // otherwise, need to change the wavfile from which music is played and move the f

    //Serial.print(stretchF
    float percentDuration = ( (float)WavFile.position() / ((float)WavHeader.DataSize - 44) ) * (float)100;
    // Serial.print("Percent duration of song before jump: "); 
    // Serial.println(percentDuration);

    WavFile.close();
    switchWavFileName();
    WavFile = SD.open(WavFileName); 

      if(WavFile==false){
        // Serial.print("Could not open ");
        // Serial.println(WavFileName);
       }
    WavFile.read((byte *) &WavHeader,44);                                                                    
    //DumpWAVHeader(&WavHeader);             // DELETE? ****                
    if(ValidWavData(&WavHeader))  {                      
      i2s_set_sample_rates(i2s_num, WavHeader.SampleRate);  
    }

    uint32_t jump_location = (( ((uint32_t)( (float)BytesReadSoFarMusicTime / speedFactorcurrent)) /2) * 2 );

    if (jump_location >= WavHeader.DataSize){
        jump_location = 44;
    }
    // make FILE pointer jump to the corresponding spot in the music
    WavFile.seek(jump_location);   

    percentDuration = ( (float)WavFile.position() / ((float)WavHeader.DataSize - 44) ) * 100;
    // Serial.print("Percent duration of song after jump: "); 
    // Serial.println(percentDuration);

    // Serial.print("WavFile.position() = ");
    // Serial.println(WavFile.position());
    // Serial.println(WavFileName);
    // Serial.println("***************** end of music data *****************\n");
    speedFactorprev = speedFactorcurrent;
  }

  if(ReadingFile)                                     // Read next chunk of data in from file if needed
  {
    BytesRead=ReadFile(Samples);                      // Read data into our memory buffer, return num bytes read in
    ReadingFile=false;                                // Switch to sending the buffer to the I2S
  }
  else
    ReadingFile=FillI2SBuffer(Samples,BytesRead);        // We keep calling this routine until it returns true, at which point
                                                      // this will swap us back to Reading the next block of data from the file.
                                                      // Reading true means it has managed to push all the data to the I2S 
                                                      // Handler, false means there still more to do and you should call this
                                                      // routine again and again until it returns true.
}

float speedFactorlist[7] = {0.8, 0.9, 1, 1.1, 1.2};
float closestValue(float x, float *list, int size) {
    if (size == 0) {
        printf("Error: Empty list.\n");
        exit(1);
    }

    float closest = list[0]; // Initialize closest to the first value in the list
    float min_diff = fabs(x - list[0]); // Initialize min_diff to the absolute difference between x and the first value

    for (int i = 1; i < size; i++) {
        float diff = fabs(x - list[i]); // Calculate the absolute difference between x and the current value in the list
        if (diff < min_diff) { // If the current difference is smaller than the minimum difference found so far
            closest = list[i]; // Update closest to the current value
            min_diff = diff; // Update min_diff to the current difference
        }
    }

    return closest; // Return the closest value
}


/**********************************************************************************************************************************************/

// Gait tracking parameters
uint16_t BNO055_SAMPLERATE_DELAY_MS = 100;
Adafruit_BNO055 bno = Adafruit_BNO055(55, 0x28, &Wire);
int steps = 0;
float t = 0;
float times_array[10];
int array_index = 0;
float time_sum = 0;
float time_avg = 0;
float time_sum_mins = 0;
float SPM = 0;

// Here we initilize all Nextion Parameters
// Format is (page number, id, name)

// page0
NexPage p0 = NexPage(0,0,"page0");
NexText p0_t2 = NexText(0,5,"t0");
NexButton p0_b0 = NexButton(0,1,"b0");
NexButton p0_b1 = NexButton(0,2,"b1");
NexButton p0_b2 = NexButton(0,3,"b2");
NexButton p0_b3 = NexButton(0,4,"b3");

// page1
NexPage p1 = NexPage(1,0,"page1");
NexText p1_t2 = NexText(1,5,"t0");
NexButton p1_b4 = NexButton(1,1,"b4");
NexButton p1_b5 = NexButton(1,2,"b5");
NexButton p1_b6 = NexButton(1,3,"b7");
NexButton p1_b7 = NexButton(1,4,"b8");

//page2
NexPage p2 = NexPage(2,0,"page2");
NexText p2_t0 = NexText(2,4,"t0");
NexNumber p2_n1 = NexNumber(2,5,"n0");
NexButton p2_b8 = NexButton(2,3,"b8");
NexButton p2_b9 = NexButton(2,2,"b9");
NexButton p2_b2 = NexButton(2,1,"b2");


//page3
NexPage p3 = NexPage(3,0,"page3");
NexText p3_t0 = NexText(3,3,"t0");
NexNumber p3_n1 = NexNumber(3,5,"n1");
NexButton p3_b10 = NexButton(3,1,"b10");
NexButton p3_b11 = NexButton(3,2,"b11");
NexButton p3_b2 = NexButton(3,4,"b2");

// Initialize nextion buttons to "listen"
// Allows them to send serial data 
NexTouch *nex_listen_list[] = {
  &p0_b0, &p0_b1, &p0_b2, &p0_b3, &p1_b4, &p1_b5, &p1_b6, &p1_b7,
  &p2_b8, &p2_b9, &p2_b2, &p3_b10, &p3_b11, &p3_b2, NULL
};

// Denote what each button does when pressed

// 60 SPM button
void p0_b0_Press(void *ptr) {
  target_SPM = 60;
  targetSPMstr = String("060"); // begin music file selection
  Serial.print("Target SPM: ");
  Serial.println(target_SPM);
}

// 75 SPM button
void p0_b1_Press(void *ptr) {
  target_SPM = 75;
  targetSPMstr = String("075"); // begin music file selection
  //WavFile = SD.open("/cher2.wav");
  Serial.print("Target SPM: ");
  Serial.println(target_SPM);
}

// 90 SPM button
void p0_b2_Press(void *ptr) {
  target_SPM = 90;
  targetSPMstr = String("090"); // begin music file selection
  Serial.print("Target SPM: ");
  Serial.println(target_SPM);
}

// 105 SPM button
void p0_b3_Press(void *ptr) {
  target_SPM = 105;
  targetSPMstr = String("105"); // begin music file selection
  Serial.print("Target SPM: ");
  Serial.println(target_SPM);
}

// Rock button
void p1_b4_Press(void *ptr) {
  genre = 0; // Rock
  gnrstr = String("rck");
  WavFileName = "/" + targetSPMstr + gnrstr + "100.wav"; // complete music file selection
  Serial.print("Rock Genre. ID: ");
  Serial.println(genre);
}

// Latin button
void p1_b5_Press(void *ptr) {
  genre = 1; // Latin
  gnrstr = String("lat");
  WavFileName = "/" + targetSPMstr + gnrstr + "100.wav"; // complete music file selection
  Serial.print("Latin Genre. ID: ");
  Serial.println(genre);
}

// Pop button
void p1_b6_Press(void *ptr) {
  genre = 2; // Pop
  gnrstr = String("pop");
  WavFileName = "/" + targetSPMstr + gnrstr + "100.wav"; // complete music file selection
  Serial.print("Pop Genre. ID: ");
  Serial.println(genre);
}

// Rap button
void p1_b7_Press(void *ptr) {
  genre = 3; // Rap
  gnrstr = String("rap");
  WavFileName = "/" + targetSPMstr + gnrstr + "100.wav"; // complete music file selection
  Serial.print("Rap Genre. ID: ");
  Serial.println(genre);
}

// Volume up button
void p2_b8_Press(void *ptr) {
  volume+=10;
  if(volume>100){
    volume = 100;
  }
  Serial.print("Volume is: ");
  Serial.println(volume);
}

// Volume down button
void p2_b9_Press(void *ptr) {
  volume-=10;
  if(volume<0){
    volume = 0;
  }
  Serial.print("Volume is: ");
  Serial.println(volume);
}

// Start button
void p2_b2_Press(void *ptr) {
  start = true; // this starts tasks 1 (gait tracking) and 3 (music processing) in RTOS
  // Serial.println("Start is now true");
  // Serial.println(start);
  speedFactorprev = 0; // this forces the music task to reset the song
  Serial.println(WavFileName);
}

// Volume up button (on stop page)
void p3_b10_Press(void *ptr) {
  volume+=10;
  if(volume>100){
    volume = 100;
  }
  Serial.print("Volume is: ");
  Serial.println(volume);
}

// Volume down button (on stop page)
void p3_b11_Press(void *ptr) {
  volume-=10;
  if(volume<0){
    volume = 0;
  }
  Serial.print("Volume is: ");
  Serial.println(volume);
}

// stop button
void p3_b2_Press(void *ptr) {
  start = false; // this stops tasks 1 (gait tracking) and 3 (music processing) in RTOS
  WavFile.close(); // close the wavefile
  // Serial.println("Stopped");
  volume = 50; // reset the volume
  // ESP.restart();
}

// Tasks - 1:BNO055 2:Nextion 3:Music processing
void task1( void * parameters){
  esp_task_wdt_init(3000, false);
  for (;;){
    // Serial will not be called until start is pressed - give ample time for this to occur
    Serial.setTimeout(1000000);
    // Start the program when start button is pressed
    if(start == 1){
      // Initialize variables
      float a_mag = 0; // magnitude of acceleration
      float euler_current = 0; // current euler angle
      float euler_last = 0; // euler angle of the previous reading
      
      /* Get a new sensor event */
      sensors_event_t orientationData , angVelocityData , linearAccelData, magnetometerData, accelerometerData, gravityData;
      imu::Vector<3> euler =
      bno.getVector(Adafruit_BNO055::VECTOR_EULER);   
      bno.getEvent(&linearAccelData, Adafruit_BNO055::VECTOR_LINEARACCEL);
      
      // Access linear acceleration values
      float x = linearAccelData.acceleration.x; 
      float y = linearAccelData.acceleration.y;
      float z = linearAccelData.acceleration.z;

      euler_current = euler.y(); // determine angle of rotation
      a_mag = pow((pow(x,2) + pow(y, 2) + pow(z,2)), 0.5); // determine magnitude of linear acceleration

      // conditionally determine if a step has been taken using linear acceleration and angle of rotation over the past two steps
      if ((a_mag>1) && (euler_current > 10) && euler_current > euler_last)
      {
        // STEP TAKEN - record time at which step was taken & increment step counter
        steps = steps + 1;
        Serial.print("Step taken: ");
        Serial.print(t);
        Serial.println(" seconds since last step.");
        times_array[array_index] = t;
        t = .7;
        int j;
        vTaskDelay(700/portTICK_PERIOD_MS);  // delay task1 by 0.7 seconds
        if (array_index == 9){ //9
          for (j = 0; j < 10; j++) //10
          {
            time_sum = time_sum + times_array[j];
            time_sum_mins = time_sum/60;
            time_avg = time_sum/20; // 20
            array_index = 0;
            SPM = 20/time_sum_mins; // 20
          }
          Serial.println("\n***************** start of bno data *****************");
          Serial.print("Average step length for the past 20 steps is: "); // 20
          Serial.println(time_avg);
          Serial.print("Total Steps: ");
          Serial.println(steps * 2);
          time_sum = 0;
          time_avg = 0;
          current_SPM = SPM;
          Serial.print("Current SPM:");
          Serial.println(current_SPM);
          Serial.println("***************** End of bno data *****************\n");

          //new
          //stretchFactorprev = stretchFactorcurrent;
          speedFactorcurrent = closestValue( target_SPM/current_SPM, speedFactorlist, 5); 
          Serial.println("\n***************** start of music data *****************");
          Serial.print("speedFactorcurrent = ");
          Serial.println(speedFactorcurrent);
          //end new
        }
        else {
        array_index = array_index + 1;
        }
      }
      else {
        t = t + 0.10;
        vTaskDelay(100 / portTICK_PERIOD_MS);
      }
    euler_last = euler_current;
    }
  }
}

void task2( void * parameters){
  for (;;){
    // task 2 - nexion user interface buttons
    nexLoop(nex_listen_list);
    vTaskDelay(1000 / portTICK_PERIOD_MS);
  }
}

void task3( void * parameters){
  // task 3 - music processing
  esp_task_wdt_init(3000, false);
  for (;;){
    Serial.setTimeout(1000000);
    if(start == 1){
        PlayWav();
    }
  }
}

void setup() {
  // put your setup code here, to run once

  Wire.begin(8,18);
  Serial.begin(115200);
  //delay(1000);
  
  //while (!Serial) delay(10);  // wait for serial port to open!

  // Serial.println("Orientation Sensor Test"); Serial.println("");

  /* Initialise the sensor */
  if (!bno.begin())
  {
    /* There was a problem detecting the BNO055 ... check your connections */
    // Serial.print("Ooops, no BNO055 detected ... Check your wiring or I2C ADDR!");
    while (1);
  }

  // NEXTION button setup
  nexInit();
  nexSerial.println("Started"); // Test
  p0_b0.attachPush(p0_b0_Press, &p0_b0);
  p0_b1.attachPush(p0_b1_Press, &p0_b1);
  p0_b2.attachPush(p0_b2_Press, &p0_b2);
  p0_b3.attachPush(p0_b3_Press, &p0_b3);
  p1_b4.attachPush(p1_b4_Press, &p1_b4);
  p1_b5.attachPush(p1_b5_Press, &p1_b5);
  p1_b6.attachPush(p1_b6_Press, &p1_b6);
  p1_b7.attachPush(p1_b7_Press, &p1_b7);
  p2_b8.attachPush(p2_b8_Press, &p2_b8);
  p2_b9.attachPush(p2_b9_Press, &p2_b9);
  p2_b2.attachPush(p2_b2_Press, &p2_b2);
  p3_b10.attachPush(p3_b10_Press, &p3_b10);
  p3_b11.attachPush(p3_b11_Press, &p3_b11);
  p3_b2.attachPush(p3_b2_Press, &p3_b2);

  // SD Card Setup
  SDCardInit();
  WavFile.close();
    i2s_driver_install(i2s_num, &i2s_config, 0, NULL);
    i2s_set_pin(i2s_num, &pin_config);


// INITIALIZE TASKS IN RTOS. 2 Cores used, with Core 1 used exclusively for music processing
  xTaskCreatePinnedToCore(
    task1,
    "task 1",
    25000, // stack size
    NULL, // parameters
    1, // priority
    NULL, // handle
    0 // core
  );

  xTaskCreatePinnedToCore(
    task2,
    "task 2",
    25000, // stack size
    NULL, // parameters
    3, // priority 3
    NULL, // handle
    0 // core
  );

  xTaskCreatePinnedToCore(
    task3,
    "task 3",
    100000, // stack size
    NULL, // parameters
    2, // priority
    NULL, // handle
    1 // core
  );
  delay(1000);

}

void loop() {
  // Nothing - using RTOS
}